//
//    File: glmesh.cc
//
//    (C) 2000-2008 Helmut Cantzler
//
//    Licensed under the terms of the Lesser General Public License.
//

#include "glmesh.h"

// Texts
#define MAIN_MENU_SIZE 14
#define DISPLAY_MENU_SIZE 6
#define ASCII_ESCAPE 27

const struct {
  const char* text;
  const char key;
} mainMenuText[MAIN_MENU_SIZE] = { {"Toggled translation/zoom (t)", 't'},
				   {"Toggled texture filter (T)", 'T'},
				   {"Toggled features (f)", 'f'},
				   {"Toggled normals (n)", 'n'},
				   {"Toggled bounding box (b)", 'b'},
				   {"Toggled lighting (l)", 'l'},
				   {"Toggled polygons (y)", 'y'},
				   {"Toggled shape colors (s)", 's'},
				   {"Lighting up (u)", 'u'},
				   {"Lighting down (d)", 'd'},
				   {"Clipping up (+)", '+'},
				   {"Clipping down (-)", '-'},
				   {"Full screen (F)", 'F'},
				   {"Quit (esc)", ASCII_ESCAPE} };

const struct {
  const char* text;
  const char key;
} displayMenuText[DISPLAY_MENU_SIZE] = { {"Texture (1)", '1'},
					 {"Solid (2)", '2'},
					 {"Frontlines (3)", '3'},
					 {"Wireframe (4)", '4'},
					 {"Points (5)", '5'},
					 {"Features (6)", '6'} };

char title[100];

GLMeshSettings settings;

float surfaceNormalLength;
int numberOfShapes;
GLuint *textureList;
GLint windowWidth = 500;
GLint windowHeight = 500;
GLint windowX = 100;
GLint windowY = 100;
int lastMouseX, lastMouseY;
enum ActionType actionMode = NONE;
GLfloat tbLastPosition[3];
GLfloat tbAxis[3];
int translationZoom;

////////////////// GLUT handlers ////////////////// 

void draw(void)
{
  // Clear the current display
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // save matrix
  glPushMatrix();
 
  // Change the modelview transformation
  glMultMatrixf((GLfloat*) settings.tbTransform);

  // Change the projection transformation
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-2.1 - settings.xShift - settings.zShift,
	   2.1 - settings.xShift + settings.zShift,
          -2.1 - settings.yShift - settings.zShift,
	   2.1 - settings.yShift + settings.zShift,
          -3.1 - settings.zShift - settings.clipping, 3.1);
  glMatrixMode(GL_MODELVIEW);

  // display main object (typically a mesh)
  switch (settings.meshDisplayMode)
    {
    case VERTICES:
      setMaterialColor(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, CYAN);
      displayPoints(settings.mesh->getVertices());
      break;

    case EDGES:
      setMaterialColor(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, CYAN);
      displayEdges(settings.mesh->getEdges());
      break;

    case SOLID:
    case TEXTURE:
      // Set polygon mode
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      displayMesh();
      break;
 
    case FRONTLINES:
      displayMeshWithStencil();
      break;

    case WIRE:
      // Set polygon mode
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      displayMesh();
      break;

    case FEATURES:
      displayFeatures();
      break;
      
      //    case FRONTLINES2:
      // draw the solid
      //      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      //      displayMesh();
      // draw the the wireframe on top of the solid
      //      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      //      glDepthFunc(GL_LEQUAL);
      //      displayMesh();
      //      glDepthFunc(GL_LESS);
      //      break;
    }

  checkForTraps("Draw Polygons", 1);
 
  // Draw features like points and edges additionally
  if (settings.displayFeatures && settings.meshDisplayMode != FEATURES)
    displayFeatures();

  checkForTraps("Draw Features", 1);

  // Draw the surface normals
  if (settings.displayNormals)
    displayNormals();

  checkForTraps("Draw Surface Normals",1);

  // Draw the bounding box
  if (settings.displayBoundingBox)
    displayBoundingBox();

  checkForTraps("Draw Bounding Box", 1);
 
  glPopMatrix();
  glFlush();
  glutSwapBuffers(); // Needed when double buffering
}

void reshape(int width, int height)
{
  int max = width > height ? width : height;

  if (settings.enableAspectRatio)
    glViewport((width-max)/2, (height-max)/2, max, max);
  else
    glViewport(0, 0, width, height);

  windowWidth = width;
  windowHeight = height;
}

void setTitle(void)
{
  strcpy(title, "Mesh Viewer -- ");

  if (translationZoom)
    strcat(title, "Translation");
  else
    strcat(title, "Zoom");
  strcat(title, " -- ");

  switch (settings.meshDisplayMode)
    {
    case TEXTURE:
      strcat(title, "Texture");
      break;
    case SOLID:
      strcat(title, "Solid");
      break;
    case FRONTLINES:
      strcat(title, "Frontlines");
      break;
    case WIRE:
      strcat(title, "Wireframe");
      break;
    case VERTICES:
      strcat(title, "Points");
      break;
    case FEATURES:
      strcat(title, "Features");
      break;
    }

  glutSetWindowTitle(title);
}

void toggleTextureFilter(void)
{
  GLint filter;
  int id;

  settings.enableTextureFilter = !settings.enableTextureFilter;
  filter = settings.enableTextureFilter ? GL_LINEAR : GL_NEAREST;

  for (int i=0; i < numberOfShapes; i++)
    {
      id=settings.mesh->getShape(i)->getTextureId();
      
      if (id != -1)
	{
	  glBindTexture(GL_TEXTURE_2D, id);
	  
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter);
	  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);
	}
    }
}

void key(unsigned char k, int x, int y)
{

  switch (k) 
    {
    case ASCII_ESCAPE: // Quit (Esc)

      glDeleteTextures(numberOfShapes, textureList);
      delete textureList;
      delete settings.mesh;

      exit(0);

    case 'T':
      toggleTextureFilter();
      break;

    case 'f':
      settings.displayFeatures=!settings.displayFeatures;
      break;

    case '1':
      settings.meshDisplayMode=TEXTURE;
      setTitle();
      break;
    case '2':
      settings.meshDisplayMode=SOLID;
      setTitle();
      break;
    case '3':
      settings.meshDisplayMode=FRONTLINES;
      setTitle();
      break;
    case '4':
      settings.meshDisplayMode=WIRE;
      setTitle();
      break;
    case '5':
      settings.meshDisplayMode=VERTICES;
      setTitle();
      break;
    case '6':
      settings.meshDisplayMode=FEATURES;
      setTitle();
      break;

    case 'c':
      settings.enableCutBackFaces=!settings.enableCutBackFaces;
      if (settings.enableCutBackFaces)
 	glEnable(GL_CULL_FACE);
      else
 	glDisable(GL_CULL_FACE);
      break;

    case 'l': // Toggled lighting (l)
      settings.enableLighting=!settings.enableLighting;
      if (settings.enableLighting) 
	glEnable(GL_LIGHTING);
      else 
	glDisable(GL_LIGHTING);
      break;

    case 'y': // Toggled polygon display (y)
      settings.displayPolygons=!settings.displayPolygons;
      break;

    case 's': // Toggled shape colors (s)
      settings.displayShapeColors=!settings.displayShapeColors;
      break;

    case 't': // Toggled translation/zoom (t)
      translationZoom=!translationZoom;
      setTitle();
      break;

    case 'F': // Full screen (f)
      glutFullScreen();
      break;

    case 'n': // Toggled normals (n)
      settings.displayNormals=!settings.displayNormals;
      break;

    case 'b': // Toggled bounding box (b)
      settings.displayBoundingBox=!settings.displayBoundingBox;
      break;

    case '=': case '+': // Clipping up (+)
      settings.clipping+=0.1;
      break;

    case '-': // Clipping down (-)
      settings.clipping-=0.1;
      break;

    case 'u': // Light up (u)
      {
	settings.lightBrightness+=settings.lightBrightness*0.1;
	GLfloat source[] = {settings.lightBrightness,
			    settings.lightBrightness,
			    settings.lightBrightness,
			    settings.lightBrightness};

	glLightfv(GL_LIGHT0, GL_DIFFUSE, source);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, source);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, source);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, source);
      }
      break;

    case 'd': // Light down (d)
      {
	settings.lightBrightness-=settings.lightBrightness/1.1*0.1;
	GLfloat source[] = {settings.lightBrightness,
			    settings.lightBrightness,
			    settings.lightBrightness,
			    settings.lightBrightness};

	glLightfv(GL_LIGHT0, GL_DIFFUSE, source);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, source);
	glLightfv(GL_LIGHT2, GL_DIFFUSE, source);
	glLightfv(GL_LIGHT3, GL_DIFFUSE, source);
      }
      break;
    }

  glutPostRedisplay();
}

void mouseButton(int button, int state, int x, int y)
{
  switch(state)
  {
  case GLUT_DOWN:
 
    lastMouseX = x;
    lastMouseY = y;
 
    switch(button)
      {
      case GLUT_LEFT_BUTTON:
        actionMode = ROTATE;
	tbPointToVector(x, y, tbLastPosition);
        break;
	//      case GLUT_RIGHT_BUTTON:
      case GLUT_MIDDLE_BUTTON:
        if (translationZoom)
          actionMode = TRANSLATE;
        else
          actionMode = ZOOM;
        break;
      }
    break;
 
  case GLUT_UP:
    actionMode = NONE;
  }
}

void mouseDrag(int x, int y)
{
  switch (actionMode)
    {
    case ZOOM:
      settings.zShift += (y-lastMouseY)/(double)windowHeight*3;
      lastMouseX = x;
      lastMouseY = y;
      glutPostRedisplay();
      break;

    case TRANSLATE:
      settings.xShift += (x-lastMouseX)/(double)windowWidth*3;
      settings.yShift -= (y-lastMouseY)/(double)windowHeight*3;
      lastMouseX = x;
      lastMouseY = y;
      glutPostRedisplay();
      break;

    case ROTATE:
      float currentPosition[3], angle, dx, dy, dz;
 
      tbPointToVector(x, y, currentPosition);
 
      dx = currentPosition[0] - tbLastPosition[0];
      dy = currentPosition[1] - tbLastPosition[1];
      dz = currentPosition[2] - tbLastPosition[2];
      angle = 90.0 * sqrt(dx * dx + dy * dy + dz * dz) * 2;
 
      tbAxis[0] = tbLastPosition[1] * currentPosition[2] -
	          tbLastPosition[2] * currentPosition[1];
      tbAxis[1] = tbLastPosition[2] * currentPosition[0] -
	          tbLastPosition[0] * currentPosition[2];
      tbAxis[2] = tbLastPosition[0] * currentPosition[1] -
	          tbLastPosition[1] * currentPosition[0];
 
      tbLastPosition[0] = currentPosition[0];
      tbLastPosition[1] = currentPosition[1];
      tbLastPosition[2] = currentPosition[2];

      glPushMatrix();
      glLoadIdentity();
      glRotatef(angle, tbAxis[0], tbAxis[1], tbAxis[2]);
      glMultMatrixf((GLfloat*) settings.tbTransform);
      glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *)settings.tbTransform);
      glPopMatrix();

      glutPostRedisplay();
      break;
    }
}

void mainMenu(int value)
{
  key(mainMenuText[value].key,0,0);
}

void displayMenu(int value)
{
  key(displayMenuText[value].key,0,0);
}

////////////////// Intialisation ////////////////// 

void openglInit(GLMeshSettings s)
{
  settings=s;

  settings.meshDisplayMode=NOTHING;
  if (settings.mesh != NULL)
    {
      if (settings.mesh->numberOfTriangles() > 0)
	settings.meshDisplayMode=SOLID;
      else if (settings.mesh->numberOfEdges() > 0)
	settings.meshDisplayMode=EDGES;
      else if (settings.mesh->numberOfVertices() > 0)
	settings.meshDisplayMode=VERTICES;

      numberOfShapes = settings.mesh->numberOfShapes();
      surfaceNormalLength = 0.02 + settings.mesh->averageTriangleSize()/10;
    }
  else
    numberOfShapes = 0;
    
  settings.featuresDisplayMode=NOTHING;
  if (settings.features != NULL)
    if (settings.features->numberOfEdges() > 0)
      settings.featuresDisplayMode=EDGES;
    else if (settings.features->numberOfVertices() > 0)
      settings.featuresDisplayMode=VERTICES;

  actionMode=NONE;

  // Initialise the display window
  glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB | GLUT_STENCIL);
  glutInitWindowSize(windowWidth, windowHeight);
  glutInitWindowPosition(windowX, windowY);
  glutCreateWindow("");
  setTitle();

  // Set background colour and clear
  glClearColor(0.0, 0.0, 0.0, 0.0);  // Black
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 
  // Initialise light source
  initLighting();
 
  // Other initialisations
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glLineWidth(1.5);
  glPointSize(1.5);
  glEnable(GL_POINT_SMOOTH);

  if (settings.enableCutBackFaces)
    glEnable(GL_CULL_FACE);
  else
    glDisable(GL_CULL_FACE);

  // Enable Smooth Shading
  //glShadeModel(GL_SMOOTH);
  // Really Nice Perspective Calculations
  //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  
  // Initialise the modelview transformation
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Texture stuff
  printf("Initialise Texture ...\n");
  initTexture();

  // Make the menu
  int i;
  glutCreateMenu(mainMenu);
  glutCreateMenu(displayMenu);
  for (i=0; i < DISPLAY_MENU_SIZE; i++)
    glutAddMenuEntry(displayMenuText[i].text, i);
  glutSetMenu(1);
  glutAddSubMenu("Display", 2);
  for (i=0; i < MAIN_MENU_SIZE; i++)
    glutAddMenuEntry(mainMenuText[i].text, i);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}

void openglStart(void)
{
  // Register callback handlers
  glutDisplayFunc(draw);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(key);
  glutMouseFunc(mouseButton);
  glutMotionFunc(mouseDrag);
  glutIdleFunc(0);
	
  // Main (infinite) loop 
  glutMainLoop();
}

int width(void)
{
  return windowWidth;
}

int height(void)
{
  return windowHeight;
}
