//
//    File: image.cc
//
//    (C) 2000-2008 Helmut Cantzler
//
//    Licensed under the terms of the Lesser General Public License.
//

#include "image.h"

extern "C"
{
  #include <jpeglib.h>
}

Image::Image()
{
  imageData=newData=NULL;
  imageHeight = imageWidth = imagePixelSize = 0;
}

Image::~Image()
{
  delete imageData;
}

// JPEG error handling //

#include <setjmp.h>

typedef struct error_mgr {
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
} *error_ptr;

void new_error_exit(j_common_ptr cinfo)
{
  error_ptr err = (error_ptr) cinfo->err;
  longjmp(err->setjmp_buffer, 1);
}

int Image::read(const char *filename)
{
  struct jpeg_decompress_struct cinfo;
  struct error_mgr jerr;
  JSAMPARRAY buffer;
  int rowSize, x, y, z, b;
  FILE *f;

  if ((f=fopen(filename, "rb")) == NULL)
    {
      //fprintf(stderr, "can't open %s\n", filename);
      return 1;
    }

  // Init error handling
  cinfo.err=jpeg_std_error(&jerr.pub);
  // Replace the standard jpeg error function
  jerr.pub.error_exit = new_error_exit;
  if (setjmp(jerr.setjmp_buffer))
    {
      jpeg_destroy_decompress(&cinfo);
      fclose(f);
      return 2;
    }

  // Init & Read header
  jpeg_create_decompress(&cinfo);
  jpeg_stdio_src(&cinfo,f);
  jpeg_read_header(&cinfo, TRUE);
  jpeg_start_decompress(&cinfo);

  imagePixelSize=cinfo.output_components;
  imageHeight=cinfo.output_height;
  imageWidth=cinfo.output_width;

  //printf("Bytes per pixel: %d\n", imagePixelSize);
  // Check if we have a RGB or gray scale image
  if (imagePixelSize != 1 && imagePixelSize != 3)
    {
      jpeg_destroy_decompress(&cinfo);
      return 2;
    }

  // Create the data structures
  delete imageData;
  imageData= new Array3D<unsigned char>(imageWidth, imageHeight,
					imagePixelSize);

  rowSize=imageWidth * imagePixelSize;
  buffer=(*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo,
				     JPOOL_IMAGE, rowSize, 1);

  // Read JPEG image line per line
  for (y=imageHeight-1; cinfo.output_scanline < cinfo.output_height; y--)
    {
      jpeg_read_scanlines(&cinfo, buffer, 1);

      for (b=x=0; x < imageWidth; x++)
	for (z=0; z < imagePixelSize; z++)
	  imageData->set(x, y, z, buffer[0][b++]);
    }

  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  fclose(f);

  return 0;
}

void Image::greyToRGB(void)
{
  if (imagePixelSize != 1)
    return;

  int x, y, z;

  newData = new Array3D<unsigned char>(imageWidth, imageHeight, 3);

  for (y=0; y < imageHeight; y++)
    for (x=0; x < imageWidth; x++)
      for (z=0; z < 3; z++)
	newData->set(x, y, z, imageData->get(x, y, 1));

  delete imageData;
  imageData=newData;

  imagePixelSize=3;
}

void Image::scale(int height, int width)
{
  float h, w;
  int y, x, z;

  // Ratio between old and new size
  h=(float)imageHeight/(float)height;
  w=(float)imageWidth/(float)width;

  // Create the data structures
  newData = new Array3D<unsigned char>(width, height, imagePixelSize);

  for (y=0; y < height; y++)
    for (x=0; x < width; x++)
      for (z=0; z < imagePixelSize; z++)
	newData->set(x, y, z, imageData->get((int)(x*w+0.5),
					     (int)(y*h+0.5), z));

  delete imageData;
  imageData=newData;

  imageHeight=height;
  imageWidth=width;
}

int Image::height(void) const
{
  return imageHeight;
}

int Image::width(void) const
{
  return imageWidth;
}

int Image::pixelSize(void) const
{
  return imagePixelSize;
}

const unsigned char* Image::data(void)
{
  return imageData->getData();
}
